package xyz.lychee.lagfixer.libs.adventure.pointer;

import java.util.Optional;
import java.util.function.Supplier;
import xyz.lychee.lagfixer.libs.adventure.builder.AbstractBuilder;
import xyz.lychee.lagfixer.libs.adventure.util.Buildable;
import xyz.lychee.lagfixer.libs.annotations.Contract;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;
import xyz.lychee.lagfixer.libs.annotations.UnknownNullability;

public interface Pointers extends Buildable<Pointers, Builder> {
   @Contract(
      pure = true
   )
   static @NotNull Pointers empty() {
      return PointersImpl.EMPTY;
   }

   @Contract(
      pure = true
   )
   @NotNull
   static @NotNull Builder builder() {
      return new PointersImpl.BuilderImpl();
   }

   <T> @NotNull Optional<T> get(final @NotNull Pointer<T> pointer);

   @Contract("_, null -> _; _, !null -> !null")
   default <T> @Nullable T getOrDefault(final @NotNull Pointer<T> pointer, final @Nullable T defaultValue) {
      return this.get(pointer).orElse(defaultValue);
   }

   default <T> @UnknownNullability T getOrDefaultFrom(final @NotNull Pointer<T> pointer, final @NotNull Supplier<? extends T> defaultValue) {
      return this.get(pointer).orElseGet(defaultValue);
   }

   <T> boolean supports(final @NotNull Pointer<T> pointer);

   public interface Builder extends AbstractBuilder<Pointers>, Buildable.Builder<Pointers> {
      @Contract("_, _ -> this")
      @NotNull
      default <T> @NotNull Builder withStatic(final @NotNull Pointer<T> pointer, final @Nullable T value) {
         return this.withDynamic(pointer, () -> {
            return value;
         });
      }

      @Contract("_, _ -> this")
      @NotNull
      <T> @NotNull Builder withDynamic(final @NotNull Pointer<T> pointer, @NotNull Supplier<@Nullable T> value);
   }
}
